#include <stdarg.h>
#include <stdio.h>
#include <string.h>

#include "wad/wadio.h"
#include "lib/safem.h"
#include "lib/log.h"
#include "texture.h"
#include "render/draw.h"
#include "banner.h"
#include "levdata.h"
#include "dsound.h"
#include "gettable.h"
#include "game.h"

typedef struct {
   Texture *icon;
   int mycount,*count;
   int timecount;
} GettableD;

static LumpNum ln=BAD_LUMPNUM;
static int ngetts=0;
static const Gettable *gett=NULL;
static GettableD *getd=NULL;

static int selection[2]={-1,-1};

void rotate_selection(int type) {
   int i=selection[type];
   int loops=-1;
   while(1) {
      i++;
      /* check bailout */
      if(loops++>ngetts) {
	 i=-1;
	 break;
      };
      /* wrap around if necessary */
      if(i>=ngetts||i<0) i=0;
      /* check selectable */
      if(type==0&&!(gett[i].flags&GK_WEPSELECT)) continue;
      if(type==1&&!(gett[i].flags&GK_SPESELECT)) continue;
      /* check we have one */
      if(getd[i].count[0]<1) continue;
      if(gett[i].bulletkind>=0&&getd[gett[i].bulletkind].count[0]<1) continue;
      /* a strange loop, but it works */
      break;
   };
   selection[type]=i;
};

void use_selection(int type,LevData *ld,int player) {
   const Gettable *gt=gett+selection[type];
   GettableD *gd=getd+selection[type];
   GettableD *gb=getd+selection[type];
   if(selection[type]<0) {
      rotate_selection(type);
      return;
   };
   if(gd->timecount) return;
   if(gt->bulletkind>=0) gb=getd+gt->bulletkind;
   if(gd->count[0]<1||gb->count[0]<gt->usenum) {
      rotate_selection(type);
      return;
   }
   else gb->count[0]-=gt->usenum;
   use_item(ld,player,gt);
   if(gt->sound>=0) play_dsound(gt->sound,0,0,0);
   gd->timecount=gt->timing;
};

/* pickup */

void pickup_gettable(int type,int num) {
   const Gettable *gt=gett+type;
   GettableD *gd=getd+type;
   if(type<0||type>=ngetts) 
      logprintf(LOG_ERROR,'G',"Strange gettable type=%d num=%d",type,num);
   else {
      int *cnt=gd->count;
      logprintf(LOG_DEBUG,'G',"pickup_gettable(%d,%d)",type,num);
      /* sound */
      if(gt->pickup_sound>=0)
	 play_dsound(gt->pickup_sound,0,0,0);
      /* send messages */
      if(*cnt==0&&(gt->flags&GK_GOT_THE))
	 game_message("YOU GOT THE %s!",gt->string);
      else if(*cnt>=gt->collect&&(gt->flags&GK_GOT_A)) {
	 game_message("THAT %s WOULD BE WASTED...",gt->string);
	 return;
      }
      else if(gt->flags&GK_GOT_A) 
	 game_message("GOT A %s.",gt->string);
      /* do the pickup */
      *cnt+=num;
      if(*cnt>gt->collect)
	 *cnt=gt->collect;
      if(gt->bulletkind>=0&&gt->bulletadd>0)
	 pickup_gettable(gt->bulletkind,gt->bulletadd);
   };
};


/* draw funcs */

static count_font=-1;

static void init_cfont(void) {
   if(count_font<0) {
      if(have_lump("IN0"))
	 count_font=init_font("SMALLIN%d",10);
      else 
	 count_font=init_font("STYSNUM%d",10);
   };
};

static void draw_count(void *fb,int c,int x,int y) {
   unsigned char buf[3]; 
   init_cfont();
   buf[0]=(c/100)%10;
   buf[1]=(c/10)%10;
   buf[2]=c%10;
   drawtext(fb,buf,3,count_font,x,y);
};

void draw_gettables(void *fb,int width,int height) {
   int i;
   const Gettable *gt=gett;
   GettableD *gd=getd;
   for(i=0;i<ngetts;i++,gt++,gd++) {
      int xo=gt->xo,yo=gt->yo;
      if(gd->icon==NULL||*gd->count<1) continue;
      if(xo<0) xo+=width-gd->icon->width;
      if(yo<0) yo+=height-gd->icon->height-(gt->collect>1?16:0);
      if((gt->flags&GK_WEPSELECT)&&i==selection[0])
	 draw_outline(fb,gd->icon,xo,yo);
      else if((gt->flags&GK_SPESELECT)&&i==selection[1])
	 draw_outline(fb,gd->icon,xo,yo);
      else
	 draw(fb,gd->icon,xo,yo);
      if(gt->collect>1) draw_count(fb,gd->count[0],xo,yo+gd->icon->height+3);
   };
};

/* do decay stuff */
void update_gettables(int ticks) {
   int i;
   const Gettable *gt=gett;
   GettableD *gd=getd;
   for(i=0;i<ngetts;i++,gt++,gd++) {
      if(gt->decay&&(gd->count[0]>0))
	 gd->count[0]-=ticks*gt->decay;
      if(gd->count[0]<0) gd->count[0]=0;
      if(gd->timecount>0) gd->timecount-=ticks;
      if(gd->timecount<0) gd->timecount=0;
   };
};

/* cheat */
void cheat_gettables(void) {
   int i;
   const Gettable *gt=gett;
   GettableD *gd=getd;
   for(i=0;i<ngetts;i++,gt++,gd++) {
      gd->timecount=0;
      gd->count[0]=gt->collect;
   };
};

/* return non-zero if we have a key of type <keytype> */
int gettable_chk_key(int keytype) {
   int i;
   const Gettable *gt=gett;
   GettableD *gd=getd;
   for(i=0;i<ngetts;i++,gt++,gd++)
      if(gd->count[0]&&gt->key==keytype) return 1;
   return 0;
};


/* init and reset funcs */

void set_gettable_special(int id,int *ptr) {
   int i;
   for(i=0;i<ngetts;i++) 
      if(gett[i].special==id) getd[i].count=ptr;
};

void init_gettables(void) {
   int i;
   /* if already inited, throw away old info first */
   if(ngetts>0) reset_gettables();
   /* load lump and allocate GetDs */
   count_font=-1;
   ln=lookup_lump("GETTABLE",NULL,NULL);
   if(!LUMPNUM_OK(ln)) return;
   ngetts=get_lump_len(ln)/sizeof(Gettable);
   logprintf(LOG_INFO,'G',"init %d gettables",ngetts);
   gett=load_lump(ln);
   getd=safe_calloc(sizeof(GettableD),ngetts);
   /* initialise */
   for(i=0;i<ngetts;i++) {
      if(gett[i].iconname[0]&&have_lump(gett[i].iconname))
	 getd[i].icon=get_misc_texture(gett[i].iconname);
      getd[i].mycount=gett[i].initial;
      getd[i].count=&getd[i].mycount;
   };
};

void reset_gettables(void) {
   if(LUMPNUM_OK(ln)) free_lump(ln);
   if(getd) safe_free(getd);
   ngetts=0;
   gett=NULL;
   getd=NULL;
};
