#include <stdarg.h>
#include <stdio.h>
#include <string.h>

#include "lib/log.h"
#include "wad/wadio.h"
#include "levdata.h"
#include "levdyn.h"
#include "lib/safem.h"

static MapLumpTypeInfo mlti[ML_NTYPES]=ML_INFO;

char *get_next_level_name(char *buf,const char *n) {
   if(n==NULL||*n==0) {
      if(LUMPNUM_OK(lookup_lump("MAP01",NULL,NULL)))
	 strcpy(buf,"MAP01");
      else
	 strcpy(buf,"E1M1");
   }
   else if(*n=='E'||*n=='e') {
      strcpy(buf,n);
      buf[3]++;
      if(buf[3]>'9') {
	 buf[3]='1';
	 buf[1]++;
      };
   }
   else {
      strcpy(buf,n);
      buf[4]++;
      if(buf[4]>'9') {
	 buf[4]='0';
	 buf[3]++;
      };
   };
   return buf;
};

void load_level(LevData *ld,const char *name,int d,int mp) {
   int i;
   logprintf(LOG_INFO,'M',"Loading %s: difficulty=%d",name,d);
   memset(ld,0,sizeof(LevData));
   strncpy(ld->name,name,8);
   ld->difficulty=d;
   ld->mplayer=mp;
   for(i=0;i<MAXPLAYERS;i++) ld->player[i]=-1;
   for(i=0;i<ML_NTYPES;i++) {
      LumpNum ln;
      int count;
      /* find and load the lump */
      /*logprintf(LOG_DEBUG,'M',"Loading %s",mlti[i].lumpname);*/
      ld->lumpnum[i]=ln=safe_lookup_lump(mlti[i].lumpname,name,NULL,LOG_FATAL);
      ld->count[i]=count=get_lump_len(ln)/mlti[i].lumpgrain;
      ld->lump[i]=load_lump(ln);
      /* if there's dynamic data for these entries, allocate and init it */
      if(mlti[i].dyngrain>0) 
	ld->dyn[i]=safe_calloc(count+mlti[i].dynextra,mlti[i].dyngrain);
      if(mlti[i].dyninit) {
	 int j;
	 char *dyn=ld->dyn[i];
	 const char *lump=ld->lump[i];
	 logprintf(LOG_DEBUG,'M',"Init %d %s",count,mlti[i].lumpname);
	 for(j=0;j<count;j++) {
	    mlti[i].dyninit(dyn,lump,ld);
	    dyn+=mlti[i].dyngrain;
	    lump+=mlti[i].lumpgrain;
	 }
      }
      /* if we allocated extras, adjust the count and throw away lump */
      if(mlti[i].dynextra) {
	 ld->count[i]+=mlti[i].dynextra;
	 free_levlumptype(ld,i);
      }
      /* OK, onto next lumptype */
   }
   /*logprintf(LOG_DEBUG,'M',"Done: %s",name);*/
};

void free_level(LevData *ld) {
   int i;
   logprintf(LOG_INFO,'M',"Freeing level data for %s",ld->name);
   for(i=0;i<ML_NTYPES;i++) {
      free_lump(ld->lumpnum[i]);
      if(ld->dyn[i]) safe_free(ld->dyn[i]);
   };
};

void free_levlumptype(LevData *ld,MapLumpType mlt) {
   logprintf(LOG_DEBUG,'M',"Pre-freeing %s for %s",mlti[mlt].lumpname,ld->name);
   free_lump(ld->lumpnum[mlt]);
   ld->lump[mlt]=NULL;
};
