
#ifndef THINGS_H
#define THINGS_H

#include "render/view.h"
#include "levdyn.h"

#define SHOOT_HEIGHT(td) ((3*(td)->proto->height)/4)

int find_empty_thing(LevData *ld);
int safe_find_empty_thing(LevData *ld);
int new_thing(LevData *ld,int proto_id,fixed x,fixed y,fixed z);
int thing_hurl(LevData *ld,int hurler,int missile_id,
	       fixed angle,fixed elevation);
int thing_shoot_single(LevData *ld,int shooter,int bullet_id,
		 fixed angle,fixed elevation);
void thing_shoot(LevData *ld,int shooter,int bullet_id,
		 fixed angle,fixed elevation,fixed arc,int num);
int thing_can_shoot_at(LevData *ld,int looker,int target);
void thing_wake_others(LevData *ld,int th);
void thing_find_enemy(LevData *ld,int th);
void thing_autoaim(LevData *ld,int th,fixed arc,fixed *angle,fixed *elev);
void thing_autotarget(LevData *ld,int th,fixed arc);
fixed line_dist(LevData *ld,int wall,fixed x,fixed y);

void player_apply_lookup(LevData *ld,int th,int lookup);

void thing_enter_phase(LevData *ld,int th,int ph);

/* return non-zero on fail */
int thing_send_sig(LevData *ld,int th,ThingSignal sig);

void thingd_apply_polar(ThingDyn *td,fixed force,fixed angle);
#define thingd_apply_forward(td,f) thingd_apply_polar(td,f,td->angle)
#define thingd_apply_sideways(td,f) thingd_apply_polar(td,f,td->angle+FIXED_HALF_PI)
void thingd_apply_torque(ThingDyn *td,fixed t);
void thingd_apply_up(ThingDyn *td,fixed f);

void thing_to_view(LevData *ld,int thing,View *view);

void thing_rotate_image(LevData *ld,int thing,fixed view_angle);

typedef enum {
   Thumped,Crossed,Activated,Damaged
} WallHitType;

/* back is set iff thing crossed from back to front, or hit back side */
void thing_hit_wall(LevData *ld,int thing,int wall,int back,WallHitType t);

/* player pressed the activate button */
void thing_activate(LevData *ld,int thing);

void update_things(LevData *ld);

void thingd_findsector(LevData *ld,ThingDyn *td);
#define thing_findsector(ld,th) thingd_findsector(ld,ldthingd(ld)+(th))
short findsector(LevData *ld,fixed x,fixed y,fixed z);

extern inline int reject_sectors(LevData *ld,int s1,int s2) {
   int bit,byte;
   if(s1<0||s2<0) return 1;
   bit=s1+(s2*ldnsectors(ld));
#if 0
   asm("bt %%eax,%1\n\t"
       "setc %%al\n\t"
       "movzbl %%al,%%eax\n\t"
       :"=a" (bit)
       :"mr" (ldreject(ld))
       );
   return bit;
#else
   byte=bit/8;
   bit%=8;
   return (ldreject(ld)[byte]>>bit)&1;
#endif
};

extern inline int reject_sector_wall(LevData *ld,int s,int w) {
   int side0=ldline(ld)[w].side[0];
   int side1=ldline(ld)[w].side[1];
   if(!reject_sectors(ld,s,ldside(ld)[side0].sector)) return 0;
   if(side1>0&&!reject_sectors(ld,s,ldside(ld)[side1].sector)) return 0;
   return 1;
};

extern inline int reject_sector_thing(LevData *ld,int s,int t) {
   return reject_sectors(ld,s,ldthingd(ld)[t].sector);
};

extern inline int reject_thing_thing(LevData *ld,int t1,int t2) {
   return reject_sector_thing(ld,ldthingd(ld)[t1].sector,t2);
};

#endif
