#include <stdarg.h>
#include <stdio.h>
#include <string.h>
#include <limits.h>

#include "lib/log.h"
#include "levdyn.h"
#include "animtex.h"
#include "dsound.h"
#include "updmap.h"
#include "things.h"
#include "game.h"

/* general helpers */

static fixed min_adjacent_ceiling(LevData *ld,int sector)  {
   int wall;
   fixed f=FIXED_MAX;
   for(wall=0;wall<ldnlines(ld);wall++)  {
      fixed g=FIXED_MAX;
      int front=ldline(ld)[wall].side[0];
      int back=ldline(ld)[wall].side[1];
      if(back<0) continue;
      front=ldside(ld)[front].sector;
      back=ldside(ld)[back].sector;
      if(front==sector) g=ldsectord(ld)[back].ceiling;
      else if(back==sector) g=ldsectord(ld)[front].ceiling;
      if(g<f) f=g;
   };
   return f;
};
static fixed min_iadjacent_floor(LevData *ld,int sector)  {
   int wall;
   fixed f=ldsectord(ld)[sector].floor;
   for(wall=0;wall<ldnlines(ld);wall++)  {
      fixed g=FIXED_MAX;
      int front=ldline(ld)[wall].side[0];
      int back=ldline(ld)[wall].side[1];
      if(back<0) continue;
      front=ldside(ld)[front].sector;
      back=ldside(ld)[back].sector;
      if(front==sector) g=ldsectord(ld)[back].floor;
      else if(back==sector) g=ldsectord(ld)[front].floor;
      if(g<f) f=g;
   };
   return f;
};

/* deal with ML_SECTOR events */
static void init_sector_event(LevData *ld,MapEvent *me) {
   fixed speed=4<<12;
   SectorDyn *sd=ldsectord(ld)+me->entity;
   switch(me->type)  {
   case(ME_OPENDOOR):
      me->term=min_adjacent_ceiling(ld,me->entity)-(4<<12);
      me->delta=speed;
      break;
   case(ME_CLOSEDOOR):
      me->term=sd->floor;
      me->delta=-speed;
      break;
   case(ME_LIFTDOWN):
      me->term=min_iadjacent_floor(ld,me->entity);
      me->delta=-speed;
      break;
   case(ME_LIFTUP):
      me->term=(ldsector(ld)[me->entity].floor)<<12;
      me->delta=speed;
      break;
   default:
      logprintf(LOG_ERROR,'M',"init_sector_event: strange eventtype (%d)",
		me->type); 
   };
};
static void do_sector_event(LevData *ld,MapEvent *me) {
   SectorDyn *sd=ldsectord(ld)+me->entity;
   /* deal with movement */
   switch(me->type)  {
   case(ME_OPENDOOR):
   case(ME_CLOSEDOOR):
      if((me->delta<0&&sd->ceiling<=me->term)||
	 (me->delta>0&&sd->ceiling>=me->term)) {
	 /* terminate the move */
	 sd->ceiling=me->term;
	 me->type=ME_NONE;
      }
      else sd->ceiling+=me->delta;
      break;
   case(ME_LIFTUP):
   case(ME_LIFTDOWN):
      if((me->delta<0&&sd->floor<=me->term)||
	 (me->delta>0&&sd->floor>=me->term)) {
	 /* terminate the move */
	 sd->floor=me->term;
	 me->type=ME_NONE;
      }
      else sd->floor+=me->delta;
      break;
   default:
      logprintf(LOG_ERROR,'M',"do_sector_event: strange eventtype (%d)",
		me->type); 
   };
   /* do start sound (only if the event didn't terminate) */
   if(me->type!=ME_NONE&&me->sound>=0) {
      if(sd->cent_r==0) 
	 logprintf(LOG_DEBUG,'M',"sector %d has no center?",me->entity);
      play_dsound(me->sound,sd->cent_x,sd->cent_y,sd->cent_r);
      me->sound=-1;
   };
   /* do stop sound (only if start sound got to play) */
   if(me->type==ME_NONE&&me->sound<0&&me->stopsound>=0) {
      if(sd->cent_r==0) 
	 logprintf(LOG_DEBUG,'M',"sector %d has no center?",me->entity);
      play_dsound(me->stopsound,sd->cent_x,sd->cent_y,sd->cent_r);
      me->stopsound=-1;
   };
};

/* deal with ML_THING events */
static void init_thing_event(LevData *ld,MapEvent *me) {
   ThingDyn *td=ldthingd(ld)+me->entity;
   me->x=td->x;
   me->y=td->y;
   me->z=td->z;
   me->angle=td->angle;
};
static void do_thing_event(LevData *ld,MapEvent *me) {
   ThingDyn *td=ldthingd(ld)+me->entity;
   switch(me->type)  {
   case(ME_NEWLEVEL):
      game_want_newlvl(0);
      me->type=ME_NONE;
      break;
   case(ME_SECRETLEVEL):
      game_want_newlvl(1);
      me->type=ME_NONE;
      break;
   case(ME_TELEPORT):
      logprintf(LOG_DEBUG,'M',"do_thing_event: teleporting %d",me->entity);
      td->x=me->x;
      td->y=me->y;
      td->z=me->z;
      td->angle=me->angle;
      thingd_findsector(ld,td);
      me->type=ME_NONE;
      break;
   default:
      logprintf(LOG_ERROR,'M',"do_thing_event: strange eventtype (%d)",
		me->type); 
   };
};

/* deal with ML_SIDE events */
static void init_side_event(LevData *ld,MapEvent *me) {
};
static void do_side_event(LevData *ld,MapEvent *me) {
   SideDyn *sd=ldsided(ld)+me->entity;
   switch(me->type)  {
   case(ME_SWITCHON):
      me->type=ME_NONE;
      if(sd->manimstate==1) return;
      logprintf(LOG_DEBUG,'M',"do_side_event: switching %d on",me->entity);
      sd->uanimstate=1;
      sd->manimstate=1;
      sd->lanimstate=1;
      break;
   case(ME_SWITCHOFF):
      me->type=ME_NONE;
      if(sd->manimstate==0) return;
      logprintf(LOG_DEBUG,'M',"do_side_event: switching %d off",me->entity);
      sd->uanimstate=0;
      sd->manimstate=0;
      sd->lanimstate=0;
      break;
   default:
      logprintf(LOG_ERROR,'M',"do_side_event: strange eventtype (%d)",
		me->type); 
   };
   /* deal with sounds */
   if(me->sound>=0) {
      play_dsound(me->sound,
		  sd->lined->cent_x,sd->lined->cent_y,
		  sd->lined->length/2);
      me->sound=-1;
   };
};

static MapEvent bogus_event;

MapEvent *insert_event(LevData *ld,
		       MapLumpType lumptype,
		       MapEventType etype,
		       int entity) {
   int i;
   MapEvent *me;
   for(i=0;i<MAXEVENTS;i++)
      if(!EVENTVALID(ld->event[i])) break;
   if(i>=MAXEVENTS) {
      logprintf(LOG_ERROR,'M',"No free event in insert_event(%d,%d,%d)",
		lumptype,etype,entity);
      return &bogus_event;
   };
   me=ld->event+i;
   /* init event */
   /*logprintf(LOG_DEBUG,'M',"insert_event(%d,%d,%d)",lumptype,etype,entity);*/
   memset(me,0,sizeof(MapEvent));
   me->ltype=lumptype;
   me->type=etype;
   me->entity=entity;
   me->start=ld->map_ticks+1;
   me->stop=me->start+65536;
   me->sound=-1;
   /* now event-specific initialisation */
   switch(me->ltype) {
   case(ML_SECTOR): init_sector_event(ld,me); break;
   case(ML_THING): init_thing_event(ld,me); break;
   case(ML_SIDE): init_side_event(ld,me); break;
   default:
      logprintf(LOG_ERROR,'M',"Strange lumptype %d in insert_event",lumptype);
   };
   return me;
};

int unqueue_event(LevData *ld,
		  MapLumpType lumptype,
		  MapEventType etype,
		  int entity) {
   int i,j=0;
   for(i=0;i<MAXEVENTS;i++) {
      if(!EVENTVALID(ld->event[i])) continue;
      if(ld->event[i].ltype==lumptype&&
	 ld->event[i].type==etype&&
	 ld->event[i].entity==entity) {
	 /*logprintf(LOG_DEBUG,'M',"event %d unqueued",i);*/
	 j++;
	 memset(ld->event+i,0,sizeof(MapEvent));
      };
   };
   return j;
};

int find_active_event(LevData *ld,
		      MapLumpType lumptype,
		      MapEventType etype,
		      int entity) {
   int i,j=0;
   for(i=0;i<MAXEVENTS;i++) {
      if(!EVENTVALID(ld->event[i])) continue;
      if(ld->event[i].start>ld->map_ticks) continue;
      if(ld->event[i].ltype==lumptype&&
	 ld->event[i].type==etype&&
	 ld->event[i].entity==entity) j++;
   };
   return j;
};

int find_event(LevData *ld,
		      MapLumpType lumptype,
		      MapEventType etype,
		      int entity) {
   int i,j=0;
   for(i=0;i<MAXEVENTS;i++) {
      if(!EVENTVALID(ld->event[i])) continue;
      if(ld->event[i].ltype==lumptype&&
	 ld->event[i].type==etype&&
	 ld->event[i].entity==entity) j++;
   };
   return j;
};


static void um_sides(LevData *ld) {
   int i;
   SideDyn *sd=ldsided(ld);   
   for(i=0;i<ldnsides(ld);i++,sd++)  {
      /* update animated textures (if any) */
      if(sd->uanim>=0) sd->utex=get_anim_texture(sd->uanim,&sd->uanimstate);
      if(sd->manim>=0) sd->mtex=get_anim_texture(sd->manim,&sd->manimstate);
      if(sd->lanim>=0) sd->ltex=get_anim_texture(sd->lanim,&sd->lanimstate);
   };
};

static void um_sectors(LevData *ld) {
   int i;
   SectorDyn *sd=ldsectord(ld);
   for(i=0;i<ldnsectors(ld);i++,sd++)  {
      /* update animated floors & ceilings (if any) */
      if(sd->fanim>=0) sd->ftex=get_anim_texture(sd->fanim,&sd->fanimstate);
      if(sd->canim>=0) sd->ctex=get_anim_texture(sd->canim,&sd->fanimstate);
   };
};

static void um_events(LevData *ld) {
   int i;
   MapEvent *me;
   for(i=0,me=ld->event;i<MAXEVENTS;i++,me++) {
      if(!EVENTVALID(*me)||me->start>ld->map_ticks) continue;
      /* has event passed its best before date? */
      if(me->stop<ld->map_ticks) {
	 logprintf(LOG_DEBUG,'M',"event %d passed its stop date",i);
	 memset(me,0,sizeof(MapEvent));
	 continue;
      };
      /* should event run now? */
      if(me->start<=ld->map_ticks) switch(me->ltype) {
      case(ML_SECTOR): do_sector_event(ld,me); break;
      case(ML_SIDE): do_side_event(ld,me); break;
      case(ML_THING): do_thing_event(ld,me); break;
      default:
	 logprintf(LOG_ERROR,'M',"Strange lumptype %d in um_events",
		   me->ltype);
      };
   };
};

void update_map(LevData *ld) {
   ld->map_ticks++;
   um_events(ld);
   um_sides(ld);
   um_sectors(ld);
};







